#ifndef __CO_MSS_BASED_SOLVER_H__
#define __CO_MSS_BASED_SOLVER_H__

#include "Solver.h"

#include <cstdlib>
#include <vector>

using namespace std;

/**
 * \class CoMSSBasedSolver
 * \brief Root interface for argumentation solvers built on CoMSS computers
 */
class CoMSSBasedSolver : public Solver {

 public:

/**
   * \fn CoMSSBasedSolver(TaskType,map<string,string>)
   * \brief initialize the problem solver given a task
   * \param task : the task (SE, EE, DC, DS) that is asked
   * \param additionalParams : the additional parameters from the command line
   */
  CoMSSBasedSolver(TaskType task, map<string,string> *additionalParams);
  
  /**
   * \fn initProblem(Attacks*,VarMap*)
   * \brief initialize a problem given attack relations and variable mappings
   * \param attacks : the attacks relation
   * \param varMap : the variable mappings (name <-> solverVariable)
   */
  virtual void initProblem(Attacks *attacks, VarMap *varMap) = 0;

  /**
   * \fn computeProblem
   * \brief launch the problem computation process
   */
  void computeProblem();

  /**
   * \fn displaySolution(ostream*)
   * \brief displays the solution of the problem to an output stream ; the printed informations are the ones required by ICCMA argumentation competition
   */
  void displaySolution(ostream *out);

  /**
   * \fn ~CoMSSBasedSolver();
   * \brief destructor
   */
  virtual ~CoMSSBasedSolver();

 protected:
  
  /**
   * \fn computeOneExtension
   * \brief Computes one extension (no max. cardinality enforcement) of the AF
   * \param assumps a set of assumptions 
   * \param resultIsModel a boolean which indicates if we expect a model (when computing extension(s)) or a decision (when computing DC or DS)
   */
  virtual void computeOneExtension(vector<int> *assumps, bool resultIsModel) = 0;

  /**
   * \fn computeOneMaximalExtension
   * \brief To compute a maximal (w.r.t. cardinality) extension from a coMSS
   * \param assumps a set of assumptions
   */
  virtual void computeOneMaximalExtension(vector<int> *assumps) = 0;

  /**
   * \fn computeAllExtensions
   * \brief To compute all the maximal (w.r.t. cardinality) extensions of the AF
   * \param assumps a set of assumptions
   */
  virtual void computeAllExtensions(vector<int> *assumps) = 0;

  /** the variable mappings */
  VarMap *varMap;

  /** the attack relation */
  Attacks *attacks;

  /** the number of arguments (different from the number of variables) */
  int nbArgs;

  /** the result of the task to be printed on the standard output */
  string result;
  
  TaskType task;
  string additionalParam;

  /**
   * \fn hasAModel
   * \brief Tells if the SAT instance has a model
   * \param assumps a set of assumptions
   * \return true if and only if the instance has a model
   */
  virtual bool hasAModel(vector<int> *assumps) = 0;

  /**
   * \fn getModel
   * \brief To obtain a model of the instance
   * \return a vector of boolean representing the model
   */
  virtual vector<bool> *getModel() = 0;

  /**
   * \fn modelToStringExtension
   * \brief To obtain a string which represents an extension from a model
   * \param model a model given as a vector of boolean
   * \return a string which represents the extension
   */
  string modelToStringExtension(vector<bool> *model);
  
};


#endif
