#ifndef __DEFAULT_SAT_BASED_SOLVER_H__
#define __DEFAULT_SAT_BASED_SOLVER_H__


#include <vector>

#include "Solver.h"
#include "SatBasedSolver.h"

#include "core/Solver.h"
#include "core/SolverTypes.h"


#define MINISAT_LBOOL_TRUE (Minisat::lbool((uint8_t)0))
#define MINISAT_LBOOL_FALSE (Minisat::lbool((uint8_t)1))
#define MINISAT_LBOOL_UNDEF (Minisat::lbool((uint8_t)2))


/** 
 * \class DefaultSatBasedSolver
 * \brief Interface used to build problem solvers where problem is coded in CNF and passed to Minisat ; the whole encoding must be processed in initProblem method redifinition
 */
class DefaultSatBasedSolver : public SatBasedSolver {

 public :
  
  /**
   * \fn DefaultSatBasedSolver(TaskType,string)
   * \brief initialize the problem solver given a task
   * \param task : the task (SE, EE, DC, DS) that is asked
   * \param additionalParams the additional parameters from the command line
   */
  DefaultSatBasedSolver(TaskType task, map<string,string> *additionalParams);

  /**
   * \fn IMinisatBasedSimpleProblemSolver::initProblem(Attacks*,VarMap*)
   * \brief initialize a problem given attack relations and variable mappings
   * \param attacks : the attacks relation
   * \param varMap : the variable mappings (name <-> solverVariable)
   */
  virtual void initProblem(Attacks *attacks, VarMap *varMap) = 0;

  /**
   * \fn ~IMinisatBasedSimpleProblemSolver();
   * \brief destructor
   */
  virtual ~DefaultSatBasedSolver();


  /**
   * \fn printClause
   * \brief To print a clause from Minisat structures
   * \param message a string to print before the clause
   * \param clause the clause to be printed
   */
  inline void printClause(string message, Minisat::vec<Minisat::Lit>& clause){
    cerr << message << " ";
    for(int i=0; i<clause.size(); ++i){
      if(sign(clause[i])) cerr << "-";
      cerr << (var(clause[i]) + 1) << " "; 
    }
    cerr << endl;
  }

  /**
   * \fn vecContains
   * \brief Tool used to know if a Minisat::vec contains a Minisat::Lit
   * \param vec the Minisat::vec
   * \param lit the Minisat::Lit
   * \return true if and only if lit belongs to vec
   */
  inline bool vecContains(Minisat::vec<Minisat::Lit>& vec, Minisat::Lit& lit){
    for(int i = 0 ; i < vec.size() ; i++){
      if(vec[i] == lit)
	return true ;
    }
    return false ;
  }

  virtual vector<bool>* getModel();
  virtual bool hasAModel(vector<int> *assumps);
  virtual void addBlockingClause();

 protected :

  /** the Minisat solver instance */
  Minisat::Solver solver;

  virtual void displayDimacs();
};


#endif
