#include "ExternalCoMSSBasedSolver.h"


ExternalCoMSSBasedSolver::ExternalCoMSSBasedSolver(SemanticName semName, TaskType task, string commandline, map<string,string> *additionalParams) : CoMSSBasedSolver(task, additionalParams) {
  this->semName = semName;
  this->commandline = commandline;
  this->ncstrs = 0;
}


void ExternalCoMSSBasedSolver::initProblem(Attacks *attacks, VarMap *varMap) {
  switch(this->semName) {
  case SEM_PREFERRED:
    break;
  default:
    cerr << "Error: cannot handle this semantic while using an external CoMSS solver" << endl ;
    exit(1);
  }

  this->varMap = varMap;
  vector<int> *vars = varMap->intVars();
  nbArgs = vars->size();

  for(vector<int>::iterator itVars = vars->begin() ; itVars != vars->end(); ++itVars) { // iterate over vars representing arguments
    vector<int> softClause ;
    softClause.push_back(*itVars);
    addClause(&softClause,false); 
  }

  vector<int> *attacksToCurrentVar;
  vector<int> binaryClause, completeClause, oppositesClause, binaryOppositesClause, completeOppositesClause;

  for(vector<int>::iterator itVars = vars->begin() ; itVars != vars->end(); ++itVars) { // iterate over vars representing arguments
    oppositesClause.push_back(-*itVars); // (-a vee -Pa)
    oppositesClause.push_back(-*itVars-nbArgs);
    addClause(&oppositesClause, true);
    oppositesClause.clear();

    completeClause.push_back(*itVars);
    completeOppositesClause.push_back(-*itVars-nbArgs);
    attacksToCurrentVar = attacks->getAttacksTo(*itVars);
    if(attacksToCurrentVar->size() == 0) {
      addClause(&completeClause, true);
      addClause(&completeOppositesClause, true);
      completeClause.clear();
      completeOppositesClause.clear();
      continue;
    }

    for(vector<int>::iterator itAttackers = attacksToCurrentVar->begin(); itAttackers != attacksToCurrentVar->end(); ++itAttackers) {
      binaryOppositesClause.push_back(*itVars+nbArgs); // Pa vee -b
      binaryOppositesClause.push_back(-*itAttackers);
      addClause(&binaryOppositesClause, true);
      binaryOppositesClause.clear();
      completeOppositesClause.push_back(*itAttackers);

      binaryClause.push_back(-*itVars); // -a vee Pb
      binaryClause.push_back(*itAttackers+nbArgs);
      addClause(&binaryClause, true);
      binaryClause.clear();
      completeClause.push_back(-*itAttackers-nbArgs);
    }
    addClause(&completeClause, true);
    addClause(&completeOppositesClause, true);
    completeClause.clear();
    completeOppositesClause.clear();
  }
  //delete vars; // to check

}


void ExternalCoMSSBasedSolver::addClause(vector<int> *cl, bool isHard) {
  strDimacs << (isHard ? WCNF_TOP : 1) << " ";
  for(vector<int>::iterator itCl = cl->begin(); itCl != cl->end(); ++itCl) {
    strDimacs << *itCl << " ";
  }
  strDimacs << "0\n";
  ++ncstrs;
}

static char *convertToCString(const std::string & s)
{
   char *pc = new char[s.size()+1];
   std::strcpy(pc, s.c_str());
   return pc; 
}


void ExternalCoMSSBasedSolver::launchExternalSolver(string cmdLine, string file) {
  pid_t pid;
  int childOut[2];

  if(-1 == pipe(childOut)) {perror("CoQuiAAS");exit(1);}
  if(-1 == (pid = fork())) {perror("CoQuiAAS");exit(1);}
  if(pid > 0) {
    wait(NULL);
    close(childOut[1]);
    FILE *childf = fdopen(childOut[0],"r");
    char buffer[1<<11];
    this->isSat = true;
    while(fgets(buffer, 1<<11, childf)) {
      if(!strncmp(buffer, "s HARD PART UNSATISFIABLE", 25)) {this->isSat = false;}
      if(!strncmp(buffer, "v ", 2)) {
	mss.resize(varMap->nVars());
	for(int i=0; i<varMap->nVars(); ++i) mss[i] = true;
	int nb = 1;
	for(char *pc = buffer+2; ; ++pc) {
	  while(*pc==' ') ++pc;
	  nb=0;
	  while(*pc!=' ' && *pc != '\n') nb=-'0'+nb*10+*pc++;
	  if(!nb) break;
	  mss[nb-1] = false;
	}
	break;
      }
    }
    fclose(childf);
    close(childOut[0]);
  } else {
    close(2);
    vector<string> tokens;
    istringstream iss(cmdLine);
    copy(istream_iterator<string>(iss),
	 istream_iterator<string>(),
	 back_inserter(tokens));
    vector<char*> ctokens;
    std::transform(tokens.begin(), tokens.end(), std::back_inserter(ctokens), convertToCString);
    char **args = new char*[ctokens.size()+1];
    for(unsigned int i=0; i<ctokens.size(); ++i) {
      if(!strcmp(ctokens[i],"FILE"))
	args[i]= strdup(file.c_str());
      else
	args[i] = ctokens[i];
    }
    args[ctokens.size()] = (char*)NULL;
    close(1);
    dup(childOut[1]);
    close(childOut[0]);
    execvp(args[0], args);
    perror("CoQuiAAS");
    exit(1);
  }
}

void ExternalCoMSSBasedSolver::computeOneMaximalExtension(vector<int> *assumps){
  char *tmpname = strdup("/tmp/tmp_CoQuiASS_ext_XXXXXX");
  mkstemp(tmpname);
  ofstream f(tmpname);
  f << "p wcnf " << varMap->nVars() << " " << (unsigned int)(ncstrs+assumps->size()) << " " << WCNF_TOP << endl;
  f << strDimacs.str();
  for(vector<int>::iterator it = assumps->begin(); it != assumps->end(); ++it)
    f << *it << "0" << endl;
  f.close();
  launchExternalSolver(commandline, string(tmpname));
  unlink(tmpname);
  free(tmpname);
  result = "[";
  bool first = true;
  for(int i = 0 ; i < varMap->nVars() ; i++){
    if(!mss[i]) continue;
    if(!first) result = result+",";
    first = false;
    result = result + varMap->getName(i+1);
  }
  result = result + "]";
}


void ExternalCoMSSBasedSolver::addMSSBlockingClause() {
  vector<int> cl;
  for(int i=0; i<varMap->nVars(); ++i) {
    if(!mss[i]) {
      cl.push_back(i+1);
    }
  }
  addClause(&cl, true);
}


void ExternalCoMSSBasedSolver::computeAllExtensions(vector<int> *assumps) {
  
  computeOneMaximalExtension(assumps);
  addMSSBlockingClause();
  string fullResult = "[" + result;
  computeOneMaximalExtension(assumps);
  while(this->isSat) {
    addMSSBlockingClause();
    fullResult = fullResult+","+result;
    computeOneMaximalExtension(assumps);
  }
  result = fullResult+"]";
}


void ExternalCoMSSBasedSolver::computeOneExtension(vector<int> *assumps, bool resultIsModel) {}
bool ExternalCoMSSBasedSolver::hasAModel(vector<int> *assumps) {return false;}
vector<bool> *ExternalCoMSSBasedSolver::getModel() {return NULL;}
