#include "ExternalSatBasedSolver.h"


ExternalSatBasedSolver::ExternalSatBasedSolver(SemanticName semName, TaskType task, string commandline, map<string,string> *additionalParams) : SatBasedSolver(task, additionalParams) {
  this->semName = semName;
  this->commandline = commandline;
  this->ncstrs = 0;
  this->isSat = false;
  this->nDimacsVars = 0;
}


void ExternalSatBasedSolver::initProblem(Attacks *attacks, VarMap *varMap) {
  this->nDimacsVars = varMap->nVars();
  switch(this->semName) {
  case SEM_STABLE:
    initProblemForStableSemantic(attacks, varMap);
    return;
  case SEM_COMPLETE:
    initProblemForCompleteSemantic(attacks, varMap);
    return;
  default:
    cerr << "Error: cannot handle this semantic while using an external SAT solver" << endl ;
    exit(1);
  }
}

void ExternalSatBasedSolver::initProblemForStableSemantic(Attacks *attacks, VarMap *varMap) {
  this->varMap = varMap;
  this->attacks = attacks ;
  vector<int> *vars = varMap->intVars();
  nbArgs = vars->size();
  vector<int> *attacksToCurrentVar;
  vector<int> binaryClause, completeClause;
  for(vector<int>::iterator itVars = vars->begin(); itVars != vars->end(); ++itVars) { // iterate over vars
    completeClause.push_back(*itVars);
    attacksToCurrentVar = attacks->getAttacksTo(*itVars);
    if(attacksToCurrentVar->size() == 0) {
      addClause(&completeClause);
      completeClause.clear();
      continue;
    }
    for(vector<int>::iterator itAttackers = attacksToCurrentVar->begin(); itAttackers != attacksToCurrentVar->end(); ++itAttackers) {
      binaryClause.push_back(-*itVars);
      binaryClause.push_back(-*itAttackers);
      addClause(&binaryClause);
      binaryClause.clear();
      completeClause.push_back(*itAttackers);
    }
    addClause(&completeClause);
    completeClause.clear();
  }
  delete vars;
}

void ExternalSatBasedSolver::initProblemForCompleteSemantic(Attacks *attacks, VarMap *varMap) {
  this->varMap = varMap;
  this->nDimacsVars = (varMap->nVars())<<1;
  vector<int> *vars = varMap->intVars();
  nbArgs = vars->size();

  vector<int> *attacksToCurrentVar;
  vector<int> binaryClause, completeClause, oppositesClause, binaryOppositesClause, completeOppositesClause;

  for(vector<int>::iterator itVars = vars->begin() ; itVars != vars->end(); ++itVars) { // iterate over vars representing arguments
    oppositesClause.push_back(-*itVars); // (-a vee -Pa)
    oppositesClause.push_back(-*itVars-nbArgs);
    addClause(&oppositesClause);
    oppositesClause.clear();

    completeClause.push_back(*itVars);
    completeOppositesClause.push_back(-*itVars-nbArgs);
    attacksToCurrentVar = attacks->getAttacksTo(*itVars);
    if(attacksToCurrentVar->size() == 0) {
      addClause(&completeClause);
      addClause(&completeOppositesClause);
      completeClause.clear();
      completeOppositesClause.clear();
      continue;
    }

    for(vector<int>::iterator itAttackers = attacksToCurrentVar->begin(); itAttackers != attacksToCurrentVar->end(); ++itAttackers) {
      binaryOppositesClause.push_back(*itVars+nbArgs); // Pa vee -b
      binaryOppositesClause.push_back(-*itAttackers);
      addClause(&binaryOppositesClause);
      binaryOppositesClause.clear();
      completeOppositesClause.push_back(*itAttackers);

      binaryClause.push_back(-*itVars); // -a vee Pb
      binaryClause.push_back(*itAttackers+nbArgs);
      addClause(&binaryClause);
      binaryClause.clear();
      completeClause.push_back(-*itAttackers-nbArgs);
    }
    addClause(&completeClause);
    addClause(&completeOppositesClause);
    completeClause.clear();
    completeOppositesClause.clear();
  }
}

void ExternalSatBasedSolver::addClause(vector<int> *cl) {
  for(vector<int>::iterator itCl = cl->begin(); itCl != cl->end(); ++itCl) {
    strDimacs << *itCl << " ";
  }
  strDimacs << "0\n";
  ++ncstrs;
}

void ExternalSatBasedSolver::displayDimacs() {
}

vector<bool>* ExternalSatBasedSolver::getModel() {
  return &model;
}

bool ExternalSatBasedSolver::hasAModel(vector<int> *assumps){
  char *tmpname = strdup("/tmp/tmp_CoQuiASS_ext_XXXXXX");
  mkstemp(tmpname);
  ofstream f(tmpname);
  f << "p cnf " << nDimacsVars << " " << (unsigned int)(ncstrs+assumps->size()) << endl;
  f << strDimacs.str();
  for(vector<int>::iterator it = assumps->begin(); it != assumps->end(); ++it)
    f << *it << "0" << endl;
  f.close();
  launchExternalSolver(commandline, string(tmpname));
  unlink(tmpname);
  free(tmpname);
  return isSat;
}

void ExternalSatBasedSolver::addBlockingClause(){
  vector<int> cl;
  int i=1;
  for(vector<bool>::iterator it = model.begin(); it != model.end(); ++it) {
    cl.push_back(*it ? -i : i);
    ++i;
  }
  addClause(&cl);
}


static char *convertToCString(const std::string & s)
{
   char *pc = new char[s.size()+1];
   std::strcpy(pc, s.c_str());
   return pc; 
}


void ExternalSatBasedSolver::launchExternalSolver(string cmdLine, string file) {
  pid_t pid;
  int childOut[2];
  
  if(-1 == pipe(childOut)) {perror("CoQuiAAS");exit(1);}
  if(-1 == (pid = fork())) {perror("CoQuiAAS");exit(1);}
  if(pid > 0) {
    wait(NULL);
    close(childOut[1]);
    FILE *childf = fdopen(childOut[0],"r");
    char buffer[1<<11];
    while(fgets(buffer, 1<<11, childf)) {
      if(!strncmp(buffer, "s SATISFIABLE", 13)) isSat=true;
      if(!strncmp(buffer, "s UNSATISFIABLE", 15)) isSat=false;
      if(!strncmp(buffer, "v ", 2)) {
	model.resize(varMap->nVars());
	int nb = 1;
	for(char *pc = buffer+2; nb; ++pc) {
	  bool minus = false;
	  while(*pc==' ') ++pc;
	  if(*pc == '-') {
	    minus = true;
	    ++pc;
	  }
	  nb=0;
	  while(*pc!=' ' && *pc != '\n') nb=-'0'+nb*10+*pc++;
	  if(nb) model[(nb>0 ? nb : -nb)-1] = !minus;
	}
      }
    }
    fclose(childf);
    close(childOut[0]);
  } else {
    vector<string> tokens;
    istringstream iss(cmdLine);
    copy(istream_iterator<string>(iss),
	 istream_iterator<string>(),
	 back_inserter(tokens));
    vector<char*> ctokens;
    std::transform(tokens.begin(), tokens.end(), std::back_inserter(ctokens), convertToCString);
    char **args = new char*[ctokens.size()+1];
    for(unsigned int i=0; i<ctokens.size(); ++i) {
      if(!strcmp(ctokens[i],"FILE"))
	args[i]= strdup(file.c_str());
      else
	args[i] = ctokens[i];
    }
    args[ctokens.size()] = (char*)NULL;
    close(1);
    dup(childOut[1]);
    close(childOut[0]);
    execvp(args[0], args);
    perror("CoQuiAAS");
    exit(1);
  }
}
