#ifndef __EXTERNAL_SAT_BASED_SOLVER_H__
#define __EXTERNAL_SAT_BASED_SOLVER_H__

#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <iterator>
#include <vector>

#include <sys/types.h>
#include <sys/wait.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <unistd.h>


#include "SatBasedSolver.h"


/**
 * \class ExternalSatBasedSolver
 * \brief A class which launches an external SAT solver (command line passed as a parameter for CoQuiAAS) to solve an argumentation problem 
 */
class ExternalSatBasedSolver : public SatBasedSolver {
  
 public:
  /**
   * \fn ExternalSatBasedSolver
   * \brief Constructor
   * \param semName : the accpetability semantics of the argument
   * \param task : the task (SE, EE, DC, DS) that is asked
   * \param commandline : the commandline used to call the external solver ; must contain FILE (replaced by a DIMACS temporaly source file by CoQuiAAS at execution time)
   * \param additionalParams the additional parameters from the command line
   */
  ExternalSatBasedSolver(SemanticName semName, TaskType task, string commandline, map<string,string> *additionalParams);
  void initProblem(Attacks *attacks, VarMap *varMap);

 protected:
  virtual void displayDimacs();
  virtual vector<bool>* getModel();
  virtual bool hasAModel(vector<int> *assumps);
  virtual void addBlockingClause();

 private:
  /**
   * \fn initProblemForStableSemantic
   * \brief Initializes the SAT instance for computing some problem with respect to stable semantics
   * \param attacks the attack relation of the AF
   * \param varMap the mapping between arguments and solver variables
   */
  void initProblemForStableSemantic(Attacks *attacks, VarMap *varMap);

  /**
   * \fn initProblemForCompleteSemantic
   * \brief Initializes the SAT instance for computing some problem with respect to complete semantics
   * \param attacks the attack relation of the AF
   * \param varMap the mapping between arguments and solver variables
   */
  void initProblemForCompleteSemantic(Attacks *attacks, VarMap *varMap);

  /**
   * \fn addClause
   * \brief To add a clause to the problem
   * \param cl the clause which must be added
   */
  void addClause(vector<int> *cl);

  /**
   * \fn launchExternalSolver
   * \brief Used to call an external software as the SAT solver
   * \param cmdLine the command line used to call the external solver
   * \param file the temporary file which contains the SAT instance to be solved by the external solver
   */
  void launchExternalSolver(string cmdLine, string file);

  /** The semantics */
  SemanticName semName;

  /** The command line to launch the external solver */
  string commandline;

  /** Buffer to write the SAT instance corresponding to the encoding */
  stringstream strDimacs;

  /** The number of constraints in the SAT instance */
  unsigned int ncstrs;

  /** Must be true if the problem is satisfiable */
  bool isSat;

  /** Vector of boolean variables to represent a model of the problem */
  vector<bool> model;

  unsigned int nDimacsVars;
};



#endif
