#include "PreferredSemanticProblemSolver.h"
#include <cstdlib>


using namespace std;


PreferredSemanticProblemSolver::PreferredSemanticProblemSolver(TaskType task, map<string,string> *additionalParams) : CoMSSBasedSolver(task, additionalParams) {}

PreferredSemanticProblemSolver::~PreferredSemanticProblemSolver() {}


// [bigwedge_{a \in A}(( a => -Pa) wedge (a <=> \bigwedge_{(b,a) in R}(Pb)))]
// wedge [bigwedge_{a:(b,a) in R}(Pa <=> (bigvee_{(b,a) in R} b))]
// 1,...,n <-> a_1,...,a_n
// n+1,...,2n <-> Pa_1,...,Pa_n

void PreferredSemanticProblemSolver::initProblem(Attacks *attacks, VarMap *varMap){

  if(this->task != TASK_CRED_INF) solver.setSoftInstance(true);
  this->varMap = varMap;
  vector<int> *vars = varMap->intVars();
  nbArgs = vars->size();
  for(int i=0; i<2*nbArgs; solver.newVar(), ++i); // init solver vars

  if(this->task != TASK_CRED_INF) {
    for(vector<int>::iterator itVars = vars->begin() ; itVars != vars->end(); ++itVars) { // iterate over vars representing arguments
      Minisat::vec<Minisat::Lit> softClause ;
      softClause.push(Minisat::mkLit(*itVars-1));
      solver.addClause_(softClause,false);
    }
  }

  vector<int> *attacksToCurrentVar;
  Minisat::vec<Minisat::Lit> binaryClause, completeClause, oppositesClause, binaryOppositesClause, completeOppositesClause;

  for(vector<int>::iterator itVars = vars->begin() ; itVars != vars->end(); ++itVars) { // iterate over vars representing arguments
    oppositesClause.push(~Minisat::mkLit(*itVars-1)); // (-a vee -Pa)
    oppositesClause.push(~Minisat::mkLit(*itVars-1+nbArgs));
    solver.addClause_(oppositesClause, true);
    oppositesClause.clear();

    completeClause.push(Minisat::mkLit(*itVars-1));
    completeOppositesClause.push(~Minisat::mkLit(*itVars-1+nbArgs));
    attacksToCurrentVar = attacks->getAttacksTo(*itVars);
    if(attacksToCurrentVar->size() == 0) {
      solver.addClause_(completeClause, true);
      solver.addClause_(completeOppositesClause, true);
      completeClause.clear();
      completeOppositesClause.clear();
      continue;
    }

    for(vector<int>::iterator itAttackers = attacksToCurrentVar->begin(); itAttackers != attacksToCurrentVar->end(); ++itAttackers) {
      binaryOppositesClause.push(Minisat::mkLit(*itVars-1+nbArgs)); // Pa vee -b
      binaryOppositesClause.push(~Minisat::mkLit(*itAttackers-1));
      solver.addClause_(binaryOppositesClause, true);
      binaryOppositesClause.clear();
      completeOppositesClause.push(Minisat::mkLit(*itAttackers-1));
      binaryClause.push(~Minisat::mkLit(*itVars-1)); // -a vee Pb
      binaryClause.push(Minisat::mkLit(*itAttackers-1+nbArgs));
      solver.addClause_(binaryClause, true);
      binaryClause.clear();
      completeClause.push(~Minisat::mkLit(*itAttackers-1+nbArgs));
    }
    solver.addClause_(completeClause, true);
    solver.addClause_(completeOppositesClause, true);
    completeClause.clear();
    completeOppositesClause.clear();
  }
  delete vars;

}

void PreferredSemanticProblemSolver::computeProblem() {
  result = "UNDEFINED";
  vector<int> *assumps = new vector<int>;
  switch(task) {
  case TASK_ONE_EXT:
    computeOneMaximalExtension(assumps);
    delete assumps;
    return;
  case TASK_ALL_EXTS:
    computeAllExtensions(assumps);
    delete assumps;
    return;
  case TASK_CRED_INF:
    if(!additionalParam.compare("")) return;
    assumps->push_back(varMap->getVar(additionalParam));
    computeOneExtension(assumps, false);
    delete assumps;
    return;
  case TASK_SKEP_INF:
    result = solver.isNotInCoMSS(varMap->getVar(additionalParam)) ? "YES" : "NO" ;
    delete assumps;
    return;
  default:
    return;
  }
}


void PreferredSemanticProblemSolver::computeOneExtension(vector<int> *assumps, bool resultIsModel) {
  bool ret = hasAModel(assumps);
  if (ret) {
    if(resultIsModel) {
      vector<bool> *model = getModel();
      result = modelToStringExtension(model);
      delete(model);
    }
    result = "YES";
    //result = resultIsModel ? modelToStringExtension(getModel()) : "YES";
  } else {
    result = "NO";
  }
}

void PreferredSemanticProblemSolver::computeOneMaximalExtension(vector<int> *assumps){
  Minisat::vec<Minisat::Lit> minisatAssumps;
  for(vector<int>::iterator itVars = assumps->begin(); itVars != assumps->end(); ++itVars) {
    minisatAssumps.push(*itVars > 0 ? Minisat::mkLit(*itVars) : ~Minisat::mkLit(*itVars));
  }
  solver.optSaveMCS = true;
  solver.verbosity = 0;
  solver.extractCoMSS();
  result = "[";
  vector<int> argsInExtension;
  for(int i = 0 ; i < solver.computedMCS.size() ; i++){
    for(int j=1; j<=nbArgs; ++j) argsInExtension.push_back(j);
    for(int j = 0 ; j < solver.computedMCS[i].size() ; j++){
      argsInExtension[solver.computedMCS[i][j]-1] = -1;
    }
    while(argsInExtension.size() > 0 && argsInExtension[argsInExtension.size()-1] < 0) argsInExtension.pop_back();
    for(unsigned int j=0; j<argsInExtension.size(); ++j) {
      if(argsInExtension[j] < 0) continue;
      result = j < argsInExtension.size()-1 ? result + varMap->getName(argsInExtension[j]) + "," : result + varMap->getName(argsInExtension[j]);
    }
  }
  result = result + "]";
}



void PreferredSemanticProblemSolver::computeAllExtensions(vector<int> *assumps) {
  Minisat::vec<Minisat::Lit> minisatAssumps;
  for(vector<int>::iterator itVars = assumps->begin(); itVars != assumps->end(); ++itVars) {
    minisatAssumps.push(*itVars > 0 ? Minisat::mkLit(*itVars) : ~Minisat::mkLit(*itVars));
  }
  solver.optSaveMCS = true;
  solver.verbosity = 0;
  solver.enumAllCoMssBlocked();
  result = "[";
  for(int i = 0 ; i < solver.computedMCS.size() ; i++){
    result = result + "[";
    vector<int> argsInExtension;
    for(int j=1; j<=nbArgs; ++j) argsInExtension.push_back(j);
    for(int j = 0 ; j < solver.computedMCS[i].size() ; j++){
      argsInExtension[solver.computedMCS[i][j]-1] = -1;
    }
    while(argsInExtension.size() > 0 && argsInExtension[argsInExtension.size()-1] < 0) argsInExtension.pop_back();
    for(unsigned int j=0; j<argsInExtension.size(); ++j) {
      if(argsInExtension[j] < 0) continue;
      result = j < argsInExtension.size()-1 ? result + varMap->getName(argsInExtension[j]) + "," : result + varMap->getName(argsInExtension[j]);
    }
    result = i < solver.computedMCS.size()-1 ? result + "]," : result + "]";
  }
  result = result + "]";
}

bool PreferredSemanticProblemSolver::hasAModel(vector<int> *assumps) {
  Minisat::vec<Minisat::Lit> lits;
  for(vector<int>::iterator itVars = assumps->begin(); itVars != assumps->end(); ++itVars) {
    lits.push(*itVars > 0 ? Minisat::mkLit(*itVars-1) : ~Minisat::mkLit(-*itVars-1));
  }
  solver.useAsCompleteSolver();
  Minisat::lbool ret = solver.solveLimited(lits);
  if(ret == MINISAT_LBOOL_TRUE)
    return true;
  else if (ret == MINISAT_LBOOL_FALSE)
    return false;
  else {
    cerr << "ERR:: Minisat returned LBOOL_UNDEF" << endl;
    return false;
  }
}

vector<bool> *PreferredSemanticProblemSolver::getModel() {
  vector<bool> *res = new vector<bool>;
  for (int i = 0; i < nbArgs; i++) {
    res->push_back(solver.model[i] == MINISAT_LBOOL_TRUE ? true : false);
  }
  return res;
}
