#ifndef __SAT_BASED_SOLVER_H__
#define __SAT_BASED_SOLVER_H__


#include <vector>

#include "Solver.h"

#include "core/Solver.h"
#include "core/SolverTypes.h"


/** 
 * \class SatBasedSolver
 * \brief Root interface for argumentation solvers built on SAT solvers
 */
class SatBasedSolver : public Solver {

 public :
  
  /**
   * \fn SatBasedSolver(TaskType,string)
   * \brief initialize the problem solver given a task
   * \param task : the task (SE, EE, DC, DS) that is asked
   * \param additionalParams the additional parameters from the command line
   */
  SatBasedSolver(TaskType task, map<string,string> *additionalParams);

  /**
   * \fn initProblem(Attacks*,VarMap*)
   * \brief initialize a problem given attack relations and variable mappings
   * \param attacks : the attacks relation
   * \param varMap : the variable mappings (name <-> solverVariable)
   */
  virtual void initProblem(Attacks *attacks, VarMap *varMap) = 0;

  /**
   * \fn computeProblem
   * \brief launch the problem computation process
   */
  void computeProblem();

  /**
   * \fn displaySolution(ostream*)
   * \brief displays the solution of the problem to an output stream ; the printed informations are the ones required by ICCMA argumentation competition
   */
  void displaySolution(ostream *out);

  /**
   * \fn ~SatBasedSolver();
   * \brief destructor
   */
  virtual ~SatBasedSolver();

 protected :

  /** the variable mappings */
  VarMap *varMap;

  /** the attack relation */
  Attacks *attacks ;

  /** the number of arguments (different from the number of variables) */
  int nbArgs;

  /** the result of the task to be printed on the standard output */
  string result;

  /**
   * \fn displayDimacs
   * \brief prints the SAT encoding of the problem in Dimacs format
   */
  virtual void displayDimacs() = 0;

  /**
   * \fn getModel
   * \brief To obtain a model of the encoding corresponding to the problem. The vector MUST be deleted after use.
   * \return a vector of bool representing the model
   */
  virtual vector<bool>* getModel() = 0;

  /**
   * \fn hasAModel
   * \brief To know if the problem has a model
   * \param assumps a set of assumptions
   * \return true if and only if the problem has a model (with respect to the assumptions)
   */
  virtual bool hasAModel(vector<int> *assumps) = 0;

  /**
   * \fn addBlockingClause
   * \brief Adds a blocking clause to the problem. Useful for model enumeration
   */
  virtual void addBlockingClause() = 0;

  /** The problem to solve */
  TaskType task;

  /** The parameter which represents the argument whom the acceptance must be decided (for DC and DS problems) */
  string additionalParam;

  /** Must be true if the user wants to print the encoding of the problem */
  bool onlyOutputProblem;

  /**
   * \fn computeOneExtension
   * \brief Computes one extension of the AF
   * \param assumps a set of assumptions 
   * \param resultIsModel a boolean which indicates if we expect a model (when computing extension(s)) or a decision (when computing DC or DS)
   */
  virtual void computeOneExtension(vector<int> *assumps, bool resultIsModel);

  /**
   * \fn computeAllExtensions
   * \brief To compute all the extensions of the AF
   * \param assumps a set of assumptions
   */
  virtual void computeAllExtensions(vector<int> *assumps);

  /**
   * \fn modelToStringExtension
   * \brief To obtain the string which represents an extension from a model of the problem
   * \param model a model of the problem
   * \return the string which represents the extension
   */
  virtual string modelToStringExtension(vector<bool> *model);
};


#endif
