#ifndef __SOLVER_FACTORY_H__
#define __SOLVER_FACTORY_H__


#include <algorithm>
#include <string>
#include <map>

#include "Solver.h"
#include "StableSemanticProblemSolver.h"
#include "CompleteSemanticProblemSolver.h"
#include "GroundedSemanticProblemSolver.h"
#include "PreferredSemanticProblemSolver.h"
#include "ExternalSatBasedSolver.h"
#include "ExternalCoMSSBasedSolver.h"

using namespace std;


/**
 * \class SolverFactory
 * \brief provides utility methods to build problem solvers given the task and the semantic that are required ; also provide some utility methods for semantic and task recognization
 */
class SolverFactory {

 public:

  /**
   * \fn getTaskType(str)
   * \brief try to recognize a task in a string
   * \param str : the string to parse
   */
  static TaskType getTaskType(string str) {
    str = str.substr(0,2);
    std::transform(str.begin(), str.end(), str.begin(), ::toupper);
    if(!str.compare("SE")) return TASK_ONE_EXT;
    if(!str.compare("EE")) return TASK_ALL_EXTS;
    if(!str.compare("DC")) return TASK_CRED_INF;
    if(!str.compare("DS")) return TASK_SKEP_INF;
    return TASK_UNDEFINED;
  }

  /**
   * \fn getSemanticName(str)
   * \brief try to recognize a semantic name in a string
   * \param str : the string to parse
   */
  static SemanticName getSemanticName(string str) {
    str = str.substr(3,2);
    std::transform(str.begin(), str.end(), str.begin(), ::toupper);
    if(!str.compare("ST")) return SEM_STABLE;
    if(!str.compare("CO")) return SEM_COMPLETE;
    if(!str.compare("GR")) return SEM_GROUNDED;
    if(!str.compare("PR")) return SEM_PREFERRED;
    return SEM_UNDEFINED;
  }

  /**
   * \fn getProblemInstanceForExternalSolver
   * \brief Return a new solver instance using an external SAT solver to handle the request
   * \param semantic the semantic of the argumentation framework
   * \param task  the task that is required
   * \param additionalParams the additional parameters from the command line
   */
  static Solver *getProblemInstanceForExternalSolver(SemanticName semantic, TaskType task, map<string,string> *additionalParams) {
    switch(semantic) {
    case SEM_STABLE:
    case SEM_COMPLETE:
      return new ExternalSatBasedSolver(semantic, task, (*additionalParams)["-externalSatSolver"], additionalParams);
    case SEM_PREFERRED:
      return new ExternalCoMSSBasedSolver(semantic, task, (*additionalParams)["-externalCoMSSSolver"], additionalParams);
    default:
      return NULL;
    }
  }

  /**
   * \fn getProblemInstance(SemanticName,TaskType,string)
   * \brief return a new problem instance given a semantic and a task ; also provide the additional parameter if necessary (case of DC or DS)
   * \param semantic : the semantic of the argumentation framework
   * \param task : the task that is required
   * \param additionalParams the additional parameters from the command line
   */
  static Solver *getProblemInstance(SemanticName semantic, TaskType task, map<string,string> *additionalParams) {
    if(task == TASK_UNDEFINED) return NULL;
    if(additionalParams->find("-externalSatSolver") != additionalParams->end() || additionalParams->find("-externalCoMSSSolver") != additionalParams->end()) {
      return getProblemInstanceForExternalSolver(semantic, task, additionalParams);
    }
    switch(semantic) {
    case SEM_STABLE:
      return new StableSemanticProblemSolver(task, additionalParams);
    case SEM_COMPLETE:
      return new CompleteSemanticProblemSolver(task, additionalParams);
    case SEM_GROUNDED:
      return new GroundedSemanticProblemSolver(task, additionalParams);
    case SEM_PREFERRED:
      return new PreferredSemanticProblemSolver(task, additionalParams);
    default:
      return NULL;
    }
  }

};


#endif
